--[[
Class: UDPIPE
Connects clients and servers with UDP.

When connecting, is uses handshaking, and when connected, it uses PING packets to discover
closed connections. When the connection is stalled, it can queue the messages, which are sent out after the connection is reestablished.

You can use a subclass instance and implement <Callbacks> to react on new packets.

Client side has only one connection (to the server), but the server side can send messages to any or all of the connected clients.
To send a message to all connected clients, use <UDPIPE.sendAll>. To send a message to the specified client, you need to call <UDPIPE.send> with a client's socket as an argument.

See also: <UDPIPE Description>

Usage:
Here are the simplest client and a server. You need a lua with luasockets to run the examples.

Client side:
(start code)
UDPIPE=require 'UDPIPE'
client=UDPIPE()
host, port='127.0.0.1', 1234
client:connect(host, port)

function client:onConnect(sock)
  self:send("HelloWorld")
end

function client:onNewMessage(msg)
  print('Got message:', msg)  
end

while true do
  client:step()
end
(end)

Server side:
(start code)
UDPIPE=require 'UDPIPE'
server=UDPIPE()
host, port='127.0.0.1', 1234
server:start(port, host)

function server:onNewMessage(msg, sock)
  self:send('Reply from server: '..msg, sock)
end

while true do
  server:step()
end
(end)

--]]

require 'class'
require 'socket'

--[[
Group: Private functions
--]]

local debug=true
--local debug=false
local orgprint=print
local print=function(...)
  if debug then orgprint(...) end
end

--[[
Function: _startUdp
Starts listening for UDP packets

Parameters:
  port - (number)
  host - (string)
  timeout - (number)
--]]
local function _startudp(port, host, timeout)
  print("Binding to host '" ..host.. "' and port " ..port.. "...")
  local sock, err=socket.udp()
  if not sock then
    return sock, err
  end

  assert(sock:setsockname(host, port))
  local ip, port = sock:getsockname()
  print("Waiting for packets on " .. ip .. ":" .. port .. "...")
  assert(sock:settimeout(timeout))
  print("Timeout set to "..timeout)
  return sock, port, ip
end

--[[
Group: Public methods

Function: constructor
Creates new UDPIPE instance.

Parameters:
 port - (number) port number
 host - (string) host
 timeout - (number) socket timeout
--]]
local M=class(function(self, port, host, timeout)
  print('UDPIPR INITIALIZED')
  self.port=port or 1111
  self.host=host or '*'
  self.timeout=0.01
  self.autoping=30 -- sec
  self.MAXCLIENTS=100

  self.sent=0
  self.received=0
  self.token='default'

  -- server mode only
  self.rcv_sockets={}
  self.serversocket=nil
  self.last_seen={}
  self.UNCONNECTED={}
  self.STALLED={}
  
  -- client mode only
  self.clientsocket=nil
  self.UNSENT={}

  self.mode='UNKNOWN' -- SERVER/CLIENT
end)

--[[
Function: setTimeout
Sets the timeout for all current and newly opened sockets.

Parameters:
  t - (number) timeout in seconds
--]]
function M:setTimeout(t)
  self.timeout=t
  if self.clientsocket then
    self.clientsocket:settimeout(t)
  end
  if self.serversocket then
    self.serversocket:settimeout(t)
  end
  if self.rcv_sockets then
    for k,v in pairs(self.rcv_sockets) do
      v:settimeout(t)
    end
  end
  return self
end

--[[
Function: setDebug
Sets the debugging.

Parameters:
 d - (bool) turn on/off debugging.
 --]]
 
function M:setDebug(d)
  if d==false then
    debug=false
  else
    debug=true
  end
  return self
end

--[[
Function: setToken
Sets the token for handshaking. Both client and server must have the same token for successful connection.

Parameters:
  t - (string) token
--]]
function M:setToken(t)
  self.token=t
  return self
end

--[[
Function: start
Starts the server.

Parameters:
  port - (number) public known server port
  host - (string) public known server IP address
  timeout - (number) [optional] socket timeout
--]]
function M:start(port, host, timeout)
  self.mode='SERVER'
  self.port=port or self.port
  self.host=host or self.host
  self.timeout=timeout or self.timeout
  local udp, port, ip=_startudp(self.port, self.host, self.timeout)
  if not udp then
    error('NO UDP - Server not started')
  end
  self.UNCONNECTED[udp]=true
  self.port=port  -- store real ip/port
  self.ip=ip
  print('Server listening on '..ip..':'..port..' for token '..self.token)
  self.rcv_sockets[#self.rcv_sockets+1]=udp  -- server socket is in the set of all receiving sockets
  self.serversocket=udp
  return self
end

--[[
Function: sendAll
Sends a message to all connected sockets (clients).

Parameters:
  msg - message
--]]
function M:sendAll(msg)
  if self.mode~='SERVER' then
    error('sendAll can be used in server mode only!')
  end
  for k,sock in ipairs(self.rcv_sockets) do
    if sock~=self.serversocket and not self.UNCONNECTED[sock] then
      sock:send(msg)
    end
  end
end

--[[
Function: send
Sends a message to a connected socket. If the socket is not connected, it buffers the data.

Parameters:
  msg - message
  sock - socket
--]]
function M:send(msg, sock)
  sock=sock or self.serversocket or self.clientsocket
  if self.mode=='UNKNOWN' then
    error('Need server or client mode')
  end
  if self.mode=='CLIENT' and not self.connected then
    self.UNSENT[#self.UNSENT+1]=msg
    print('Queued message: ', msg)
  else
    if self.UNCONNECTED[sock] then
      print('*** NOT SENDING - UNCONNECTED SOCKET Queuing for the future.')
      self.UNSENT[#self.UNSENT+1]=msg
    else
      if not sock then
        if self.mode=='CLIENT' and not self.clientsocket and not self.reconnect then
          print('Not sending - was connection closed?')
          return
        end
        error('Mode-client but no socket!, mode, sock='..self.mode..','..tostring(sock)..','..tostring(self.clientsocket)..' msg='..msg)
        
      end
      local res, err=sock:send(msg)
      if not res then
        print('Error sending message:', err, msg)
      else
        self.sent=self.sent+1
      end
    end
  end
end

--[[
Function: getSentPackets
Returns the number of packets sent out.

Returns:
  totalSent - (number) number of packets sent.
--]]
  
function M:getSentPackets()
  return self.sent
end

--[[
Function: getReceivedPackets
Returns the number of packets received.

Returns:
  totalReceived - (number) number of packets received.
--]]
  
function M:getReceivedPackets()
  return self.received
end

--[[
Function: ping
Sends a ping message to server (in client mode) or all connected sockets (in server mode).
--]]
function M:ping(sock)
  if self.serversocket then
    if sock then
      self:send('PING', sock)
    else
      self:sendAll('PING')
    end
  end
  if self.clientsocket then
    self:send('PING')
  end
end

--[[
Function: step
Keeps the client/server running.

Parameters:
  dt - delta time
--]]

function M:step(dt)
  dt=dt or 0.1
  if self.serversocket then
    local readable,writable,e=socket.select(self.rcv_sockets, nil, self.timeout)

    for src,udp in ipairs(readable) do
      local dgram, ip, port
      repeat
        if udp==self.serversocket or self.UNCONNECTED[udp] then
          dgram, ip, port = udp:receivefrom()
        else
          dgram=udp:receive()
        end

        if dgram then
          if udp==self.serversocket then
            self:_onNewClient(dgram, ip, port)
          else
            self:_onNewMessage(dgram, udp, ip, port)
          end
          if self.STALLED[udp] then
            self.STALLED[udp]=nil
            self:onUnstalled(udp)
          end
        end
      until dgram==nil
    end
    if #readable==0 then
      self:_onTimeout()
    end
    return #readable
  end

  if self.clientsocket then
    if self.connecting then
      if self.connectingdelay==0 then
        self.clientsocket:sendto('HELLO '..self.token, self.host, self.port)
      end
      self.connectingdelay=self.connectingdelay+dt
      if self.connectingdelay>1 then -- every 1 second
        self.connectingdelay=0
      end
    end -- connecting
    
    local res, err=self.clientsocket:receive()
    if res then
      self:_onNewMessage(res, self.clientsocket)
      self._apelapsed=0 -- reset autoping timer
    else
      if err=='connection refused' then
        if self.reconnect then
          if self.connected then
            self.connected=nil
            self:_unsetPeer(self.clientsocket)
            print('REMOTE SERVER HAS CLOSED CONNECTION')
            self:onDisconnect(self.clientsocket)
            self.connecting=true
            self.connectingdelay=0
          else
            print('Not reconnecting')
          end
        else
          print('Internal error :( ')
        end
      else
        -- timed out
        self:_onTimeout()
      end
      -- AutoPing
      if self.connected then
        self._apelapsed=(self._apelapsed or 0)+dt
        if self._apelapsed>=self.autoping then
          self:ping()
          self._apelapsed=0
        end
      end
    end
  end
end

--[[
Function: close
Close the connection.

In case of client, it also disables automatic reconnection. Otherwise if the connection was broken by the server,  the client will be reconnecting to the server.

Parameters:
  skt - socket
--]]
function M:close(skt)
  if self.clientsocket then
    self:send('CLOSED')
    self.clientsocket:close()
    self.clientsocket=nil
    self.reconnect=nil
  end
  if self.serversocket then
    if skt then
      if not self.UNCONNECTED[skt] then
        skt:send('CLOSED')
      end
      skt:close()
      for k,v in ipairs(self.rcv_sockets) do
        if v==skt then
          table.remove(self.rcv_sockets, k)
          break
        end
      end
      self.last_seen[skt]=nil
      self.UNCONNECTED[skt]=nil
    else
      self:sendAll('CLOSED')
      for k,v in ipairs(self.rcv_sockets) do
        v:close()
      end
      sel.serversocket=nil
      self.rcv_sockets={}
      self.UNCONNECTED={}
      self.last_seen={}
    end
  end
  self:onClose(skt)
  return self
end

--[[
Function: connect
Connects as a client to the server.

Parameters:
  host - server address
  port - server port
  timeout - timeout
--]]
function M:connect(host, port, timeout)
  print('Connect:', host, port, timeout)
  self.mode='CLIENT'
  self.port=port or self.port --remote
  self.host=host or self.host --remote
  self.timeout=timeout or self.timeout
  local udp, lport, ip=_startudp(0, '*', self.timeout)
  self.UNCONNECTED[udp]=true
  self.clientsocket=udp
  self.connecting=true
  self.connectingdelay=0
  self.reconnect=true
  return self
end

--[[
Group: Private methods
--]]
-- Server: no info from client
-- Client: np info from server
--[[
Function: _onTimeout
Called when no packet is received in some amount of time.
--]]
function M:_onTimeout()
  local t=os.time()
  for sock,ts in pairs(self.last_seen) do
    if t-ts>30 and self.serversocket then  -- Disconnected if no packet for 30 secs
      self:onDisconnect(sock)
      self:close(sock)
    elseif t-ts>10 and not self.STALLED[sock] then -- Stalled if no packet for 10 secs
      self.STALLED[sock]=true
      self:ping(sock)
      self:onStalled(sock)
    end
  end
end

--[[
Function: _onNewClient
Callback called when new client connection is detected. It opens the new private connected UDP socket for the client,
or calls onMaxClient if the limit is exceeded. Finally it calls <onNewClient>().

Parameters:
  msg - (string) message
  ip - (string) IP address
  port - (number) port
--]]
function M:_onNewClient(msg, ip, port)
  print('Got new client', ip, port, msg)
  if msg~='HELLO '..self.token then
    self.serversocket:sendto('GO AWAY!', ip, port) -- wrong token
    self:onBadToken(msg, ip, port)
    return
  end
  local udp, portc, ipc=_startudp(0, self.host, self.timeout)
  if not udp or #self.rcv_sockets>(self.MAXCLIENTS or 1000) then
    self:onMaxClients(msg, ip, port)  -- too many clients (flooding?)
    return
  end
  self.UNCONNECTED[udp]=true
  if ipc=='0.0.0.0' then ipc=self.host end
  self.rcv_sockets[#self.rcv_sockets+1]=udp
  self.connected=true
  self.serversocket:sendto('CONNECT '..ipc..' '..portc, ip, port)  -- Start handshaking
  self.last_seen[udp]=os.time()
  self:onNewClient(msg, ip, port, ipc, portc)
end


--[[
Function: _setPeer
Sets the peer for the socket, makes socket connected.

Parameters:
  sock - network socket
  rip - remote IP
  rport - remote port
--]]
function M:_setPeer(sock, rip, rport)
  sock:setpeername(rip, rport)
  self.UNCONNECTED[sock]=nil
end

--[[
Function: _unsetPeer
Unsets the peer for the socket, makes socket unconnected.

Parameters:
  sock - network socket
--]]
function M:_unsetPeer(sock)
  sock:setpeername('*')
  self.UNCONNECTED[sock]=true
end

--[[
Function: _onNewMessage
Called when new packed arrives on connected socket.

Parameters:
  msg - (string) message
  sock - (userdata) socket
  ip - IP
  port - port
--]]
function M:_onNewMessage(msg, sock, ip, port)
  self.received=self.received+1
  if msg=='CONNECT' then
    -- server
    -- not yet connected
    self:_setPeer(sock, ip, port)
    sock:send('CONNECTED')
    self:onConnect(sock, ip, port)
  elseif msg=='PING' and not self.UNCONNECTED[sock] then
    sock:send('PONG')
    self:onPing(sock)
  elseif msg=='PONG' and not self.UNCONNECTED[sock] then
    self:onPong(sock)
  elseif msg=='CLOSE' then
    if self.mode=='SERVER' then
      self:close(sock)
    else
      self:close()
    end
  elseif msg=='CONNECTED' then
    -- client
    self:onConnect(self.clientsocket) -- client connected
    -- skip
  else
    local rip, rport=msg:match('CONNECT ([%w%.]+) (%d+)')
    -- client
    if rip then
      sock:sendto('CONNECT', rip, rport)
      self:_setPeer(sock, rip, rport)
      self.connected=true
      self.connecting=nil
      if #self.UNSENT>0 then
        print('Sending queued messages')
        local T=self.UNSENT -- FIXME - copy table!
        for k,msg in ipairs(T) do
          self:send(msg, self.clientsocket)
        end
        self.UNSENT={}
      end
      self:onConnecting(self.clientsocket, rip, rport) -- client connecting to the final port
    else
      self:onNewMessage(msg, sock)
    end
  end
  self.last_seen[sock]=os.time()
end

--[[
Group: Callbacks

Function: onNewClient
Called when new client is connected.

Parameters:
  msg - (string)
   ip - (string) server IP
   port - (number) server port
   ipc - (string) client IP
   portc - (number) client port
--]]
function M:onNewClient(msg, ip, port, ipc, portc)
end

--[[
Function: onNewMessage
Called when new message arrives.

Parameters:
  msg - message
  sock - socket
--]]
  
function M:onNewMessage(msg, sock)
end


--[[
Function: onDisconnect
Called when a socket is disconnected.

Parameters:
  sock - client socket (in case of server code) or server socket.

The sock argument makes sense only in server code, as client is always connected (only) to the server.
--]]
function M:onDisconnect(sock)
  --print('DISCONNECTED!')
end

--[[
Function: onConnect
Called when client connected to the Server, or a new client has connected.

Parameters:
  sock - socker
  ip - client ip (in server mode only)
  port - client port (in server mode only)
--]]
function M:onConnect(sock, ip, port)
  --print('CONNECTED TO SERVER!')
end

--[[
Function: onClose
Called when client or server closed the connection.

Parameters:
  sock - socker
--]]
function M:onClose(sock)
end

--[[
Function: onConnecting
Called when client is connecting to the Server on a final port.

Parameters:
  sock - socker
  ip - remote server ip
  port - remote server port
--]]
function M:onConnecting(sock, ip, port)
  --print('CONNECTING TO SERVER!')
end

--[[
Function: onStalled
Called when the connection lags.

Parameters:
  sock - socket
--]]
  
function M:onStalled(sock)
  --print('WARN stalled socket!', tostring(sock))
end

--[[
Function: onUnstalled
Connection is no longer stalled.

Parameters:
  sock - socket
--]]
  
function M:onUnstalled(sock)
  --print('WARN stalled socket!', tostring(sock))
end

--[[
Function: onMaxClients
Called when too much client connections is open.

Parameters:
  msg - last client message
  ip - ip
  port - port
--]]
  
function M:onMaxClients(msg, ip, port)
  --print('Max clients reached:', #self.rcv_sockets,  msg, ip, port)
end
--[[
Function: onPong
Called when PONG packet is received.

Parameters:
  sock - (userdata) socket
--]]
function M:onPong(sock)
end

--[[
Function: onPing
Called when PING packet is received.

Parameters:
  sock - (userdata) socket
--]]
function M:onPing(sock)
end

--[[
Function: onBadToken
Called when bad token is received in HELLO packet.
The server expects a "HELLO <token>" as a first message from new client. See <setToken>.

Parameters:
  msg - (string) original packet payload
  ip - (string) client ip
  port - (number) client port
--]]
function M:onBadToken(msg, ip, port)
end

function M:__tostring()
  return string.format('<UDP %s %s:%s>', self.mode, self.host, self.port)
end

return M
